from mulgrids import *
import string

def quick_enthalpy(t_or_p,ph='liq'):
    """ Return enthalpy J/kg ('liq', 'vap', or 'dif') of water at specified 
        temperature (<=500.0 in degC) or pressu (>500.0 in Pa) """
    import t2thermo
    def enth(t,p,f):
        d,u = f(t,p)
        return u + p/d
    def hlhs(t,p):
        return enth(t,p,t2thermo.cowat), enth(t,p,t2thermo.supst)
    def sat_tp(t_or_p):
        if t_or_p > 500.0:
            return t2thermo.tsat(t_or_p), t_or_p
        else:
            return t_or_p, t2thermo.sat(t_or_p)
    (hl,hs) = hlhs(*sat_tp(t_or_p))
    # xxxx
    return {'liq': hl,'vap': hs,'dif': hs-hl}[ph]

class RelativePermeability:
    """ Class to calculate linear relative permeability, this is used in
        TOUGH2 for the linear relative permeability model.  The class has a
        single method, which takes the saturation and returns the relative
        permeability. """
    def __init__(self, wj_object=None):
        if wj_object is None:
            wj_object = {
                "type": "linear",
                "liquid": [0.0, 1.0],
                "vapour": [0.0, 1.0]
            }
        self.setting = wj_object
        self.func = {
            "linear": self.linear,
        }

    def calc(self, vapour_saturation):
        """ return (kr_liquid, kr_vapour) """
        return self.func[self.setting['type']](vapour_saturation)

    def linear(self, vapour_saturation):
        liq_limits = self.setting['liquid']
        vap_limits = self.setting['vapour']
        liquid_saturation = 1.0 - vapour_saturation
        kr_liq = np.interp(liquid_saturation, liq_limits, [0., 1.], left=0.0, right=1.0)
        kr_vap = np.interp(vapour_saturation, vap_limits, [0., 1.], left=0.0, right=1.0)
        return kr_liq, kr_vap

def flowing_enthalpy(lst, wj, block):
    """ calculate flowing enthalpy of a given block if producing from waiwera h5

    NOTE use values from waiwera h5 at current time
    """
    phases = ['liquid', 'vapour']
    rp = RelativePermeability(wj['rock']['relative_permeability'])
    vapour_saturation = lst.element[block][f'fluid_vapour_saturation']
    rel_perm = rp.calc(vapour_saturation)
    fluid = lst.element[block]
    # mobility
    mobility, sum_mobility = {}, 0.0
    for iph, phase in enumerate(phases):
        density = fluid[f'fluid_{phase}_density']
        viscosity = fluid[f'fluid_{phase}_viscosity']
        if viscosity == 0.0:
            mobility[phase] = 0.0
        else:
            mobility[phase] = rel_perm[iph] * density / viscosity
        sum_mobility += mobility[phase]
    # flow fraction
    flow_frac = {}
    for phase in phases:
        flow_frac[phase] = mobility[phase] / sum_mobility
    # enthalpy
    enthalpy = 0.0
    for phase in phases:
        enthalpy += flow_frac[phase] * fluid[f'fluid_{phase}_specific_enthalpy']
    # breakpoint()
    return enthalpy

def bottomhole_pressure(depth, temperature=20.0, whp=1.0e5, division=100, 
                        min_depth_interval=10.0):
    """ Calculate the pressure at the bottom of a well, given the wellhead
        pressure (whp) in pascal, temperature in degC, and division in m. The
        pressure is calculated as: pressure = whp + rho * G * depth where rho is
        the density of water at the given temperature, G is the gravitational
        acceleration (9.81 m/s^2), and depth is the depth of the well divided by
        division.

        What tis code does is then divide the depth into inv=tervals.  Then the
        pressure is accumulated, this allows the density calculation to use the
        actual pressure at the depth.  For liquid water, it is quite
        incompressible, so the result won't be that different to the easy
        rho*G*depth method.        
    """
    import t2thermo
    # divide well depth to intervals, use larger of the min interval and division
    interval = max(min_depth_interval, depth/float(division))
    d = 0.0
    pressure = whp
    while d < depth:
        rho, u = t2thermo.cowat(temperature, pressure)
        pressure += rho * 9.81 * interval
        d += interval
    rho, u = t2thermo.cowat(temperature, pressure)
    pressure += rho * 9.81 * (depth - d)
    return pressure

def block_depth(block, geo):
    """ Return the depth of a block in a geo object, this is the depth of the
        centre of the block.  This also work for Waiwera cell if block is an
        integer number.
    """
    if isinstance(block, int):
        block = geo.block_name_list(block + geo.num_atmosphere_blocks)
    lay, col = geo.layer_name(block), geo.column_name(block)
    return geo.column[col].surface - geo.block_centre(lay, col)[2]

def xyz2fit(fn):
    data = np.fromfile(fn,sep=" ")
    nrow = np.size(data) / 3
    data= data.reshape(nrow,3)
    return data

def find_all_cols_below(geo,level,surfer_file=None):
    if surfer_file is not None: outfile = file(surfer_file,'w')
    cols = []
    for col in geo.columnlist:
        surf = col.surface
        if surf < level:
            #print(col.name, str(surf))
            cols.append([col.centre[0], col.centre[1], surf, col.name])
            if surfer_file is not None: outfile.write(
                str(col.centre[0])+' '+str(col.centre[1])+' '+str(surf)+' '+
                col.name + '\n')
    if surfer_file is not None:
        outfile.close()
        print(' -- file: ', surfer_file, ' is written.')

def t2_strict_name(n):
    """ convert any name into the common TOUGH style 5 character name """
    import string
    def make_number(c):
        """ change a character into a single number, mostly 'A' or 'a' will 
            become a '1' """
        if len(c.strip()) == 0: return ' '
        d = c.lower().strip()[:1]
        i = string.ascii_lowercase.find(d)+1
        if i == 0: return ' '
        else: return str(i%10)
    newn = list(n.strip())
    if len(newn) <= 3:
        return "".join(newn).strip()[:5].ljust(5)
    for i in range(3,len(newn)):
        if newn[i] not in '0123456789':
            newn[i] = make_number(newn[i])
    return "".join(newn).strip()[:5].ljust(5)

def is_leap_year(y):
    if   int(y)%400 == 0: return True
    elif int(y)%100 == 0: return False
    elif int(y)%4   == 0: return True
    else:                 return False

def days_in_month(month, leap_year=False):
    if leap_year:
        d_month = [  31 , 29 , 31 , 30 , 31 , 30 , 31 , 31 , 30 , 31 , 30 , 31  ]
    else:
        d_month = [  31 , 28 , 31 , 30 , 31 , 30 , 31 , 31 , 30 , 31 , 30 , 31  ]
    return d_month[month - 1]

def date2str(d,m,y):
    ds, ms, ys = str(d), str(m), str(y)
    while len(ds) < 2: ds = '0'+ds
    while len(ms) < 2: ms = '0'+ms
    while len(ys) < 4: ys = '0'+ys
    return ds+'/'+ms+'/'+ys

def date2num(enddate):
            
    d,m,y = enddate.split('/')
    months    = [ '01','02','03','04','05','06','07','08','09','10','11','12' ]
    d_month   = [  31 , 28 , 31 , 30 , 31 , 30 , 31 , 31 , 30 , 31 , 30 , 31  ]
    d_month_l = [  31 , 29 , 31 , 30 , 31 , 30 , 31 , 31 , 30 , 31 , 30 , 31  ]
    acum_ds   = [sum(d_month[:i]) for i in range(12)]
    acum_ds_l = [sum(d_month[:i]) for i in range(12)]
    ad_m   = dict(zip(months, acum_ds)) # accumulated days before this month
    ad_m_l = dict(zip(months, acum_ds_l)) # accumulated days before this month
    ds_m   = dict(zip(months, d_month)) # maximum days in this month
    ds_m_l = dict(zip(months, d_month_l)) # maximum days in this month
    # check and process d and m, this depends on data
    if m not in months:
        print(' Error, unable to convert ', enddate, ' to numeric format. check month.')
        sys.exit()
    if is_leap_year:
        if int(d) not in range(1,ds_m_l[m]+1):
            print(' Error, unable to convert ', enddate, ' to numeric format. check day.')
            sys.exit()
        num = float(y) + ((float(d) + float(ad_m_l[m]))/float(sum(d_month_l)))
    else:
        if int(d) not in range(1,ds_m[m]+1):
            print(' Error, unable to convert ', enddate, ' to numeric format. check day.')
            sys.exit()
        num = float(y) + ((float(d) + float(ad_m[m]))/float(sum(d_month)))
    return num


def identifier(x, chars='abcdefghijklmnopqrstuvwxyz', width=5):
    """ creates a character-based unique identifier from a given integer,
        both chars and width are customisable, output is space filled and
        right aligned """
    output = []
    base = len(chars)
    while x:
        output.append(chars[x % base])
        x /= base
    final = ''.join(reversed(output))
    if len(final) > width:
        raise Exception('identifier() failed, not enough width.')
    return ('%' + str(width) + 's') % final




""" handle GMF's boundary condition



"""

def cell2block(geo, cell):
    """ returns mulgrid block name (str) of a cell id (int) """
    return geo.block_name_list[geo.num_atmosphere_blocks:][cell]

def update_waiwera_model_heat_bc(geo, wjson, bcjson):
    """ bcjson is loaded from the GMF modifier _boundary_conditions.json
    """
    # heat component
    if wjson['eos']['name'] == 'wae':
        heat_components = [3, 'energy']
    else:
        raise NotImplementedError(f"(gimu)    EOS {wjson['eos']['name']} not implemented")


    if 'Heat' not in bcjson:
        return
    heat_bc = bcjson['Heat']
    default_flux = float(heat_bc['Default']['Heat_Flux'])
    count = 0
    for src in wjson['source']:
        if 'component' in src:
            if src['component'] in heat_components:
                block_name = cell2block(geo, src['cell'])
                column_name = geo.column_name(block_name)
                area = geo.column[column_name].area
                flux = default_flux
                for zone_name, zone in heat_bc.get('Zones', {}).items():
                    if column_name in zone['Columns']:
                        flux = float(zone['Heat_Flux'])
                        break
                src['rate'] = area * flux
                count += 1
                # print(f"Updated {src['name']} heatflow")
    print(f"(gimu)    Updated heatflow in {count} sources")


def update_aut2_model_heat_bc(geo, dat, bcjson):
    if 'Heat' not in bcjson:
        return
    heat_bc = bcjson['Heat']
    default_flux = float(heat_bc['Default']['Heat_Flux'])
    count = 0
    for gen in dat.generatorlist:
        if gen.type == 'HEAT':
            block_name = gen.block
            column_name = geo.column_name(block_name)
            area = geo.column[column_name].area
            flux = default_flux
            for zone_name, zone in heat_bc.get('Zones', {}).items():
                if column_name in zone['Columns']:
                    flux = float(zone['Heat_Flux'])
                    break
            gen.gx = area * flux
            count += 1
            # print(f"Updated {gen.name} heatflow")
    print(f"(gimu)    Updated heatflow in {count} generators")






import json
import sys
from pathlib import Path

def main():
    """
    Duplicates a JSON file, adding a postfix to the filename and updating
    an internal value.
    """
    if len(sys.argv) != 3:
        print("Usage: make_waiwera_copy <input.json> <postfix>")
        sys.exit(1)

    input_path = Path(sys.argv[1])
    postfix = sys.argv[2]

    if not input_path.exists() or input_path.suffix != '.json':
        print(f"Error: Input file '{input_path}' does not exist or is not a .json file.")
        sys.exit(1)

    # Construct output path
    output_path = input_path.with_name(f"{input_path.stem}{postfix}{input_path.suffix}")

    # Read and modify JSON content
    with open(input_path, 'r') as f:
        data = json.load(f)

    if "output" in data and "filename" in data["output"]:
        h5_path = Path(data["output"]["filename"])
        new_h5_filename = f"{h5_path.stem}{postfix}{h5_path.suffix}"
        data["output"]["filename"] = new_h5_filename
    else:
        print("Warning: ['output']['filename'] key not found in the JSON file. No change made to content.")

    # Write new JSON file
    with open(output_path, 'w') as f:
        json.dump(data, f, indent=2)

    print(f"Successfully created '{output_path}'")

if __name__ == "__main__":
    main()



import importlib.metadata
from pathlib import Path
from pprint import pprint
import sys

try:
    import tomllib  # Python ≥3.11
except ImportError:
    import tomli as tomllib  # Python <3.11

def get_distribution():
    top_pkg = __package__.split('.')[0] if __package__ else None
    if top_pkg:
        try:
            return importlib.metadata.distribution(top_pkg)
        except importlib.metadata.PackageNotFoundError:
            pass
    return None

def get_project_meta():
    """ return pyproject.toml [project] """
    current_file = Path(__file__).resolve()
    for parent in [current_file.parent] + list(current_file.parents):
        pyproject = parent / "pyproject.toml"
        if pyproject.exists():
            with open(pyproject, "rb") as f:
                data = tomllib.load(f)
            return data.get("project", {})
    return None

def show_help():
    dist = get_distribution()
    name = dist.metadata['Name']
    version = dist.metadata['Version']
    description = dist.metadata['Summary']
    commands = [ep.name for ep in dist.entry_points]

    # meta = get_project_meta()
    # name = meta['name']
    # description = meta['description']

    msg = '\n'.join([
        f"",
        f"    {name} - {description}",
        f"",
        f"    Version: {version}",
        f"",
        f"    Available commands:",
    ] + ["        " + cmd for cmd in sorted(commands)] + [
        f"",
        f"    e.g.:",
        f"        save2incon model_1.save model_2.incon",
        f"",
    ])
    print(msg)




""" handle GMF's modifier files that update model inputs with meta- parameters

- reads rocktype JSON and update model input
- reads upflow JSON and update model input
"""

import json
import os
import sys

def compute_rocktypes(rocktype_json, warn=True):
    """ compute the derived rocktype properties and update in place

    if warn, will raise exception when rules are not followed:
        - along modifier >= 1.0
        - across modifier <= 1.0

    Example entry in rocktype_json:
    - base rocktype
        "A0000": {
          "permeability": [1e-11, 1e-11, 1e-12 ],
          "porosity": [0.25 ]
        },

    - fault modified rocktype
        "AA000": {
          "base": ["A0000"],
          "fault direction": [1],
          "modifier": {
            "along": [1],
            "across": [1],
            "vertical": [1],
            "porosity": [1]
          },
          "permeability": [1e-11, 1e-11, 1e-12],
          "porosity": [0.25]
        },

    - fault interesection rocktype
        "SVU11": {
          "base": ["S0011"],
          "modifier": {
            "k1": [2.5],
            "k2": [0.4],
            "k3": [5],
            "porosity": [1]
          },
          "permeability": [2.5000000000000004e-15, 4.0000000000000004e-16, 5.000000000000001e-15],
          "porosity": [0.1]
        }

    """
    for rt_name, rt in rocktype_json.items():
        if 'base' not in rt:
            # base rocktype
            pass
        else:
            if 'fault direction' in rt:
                fault_dir = rt['fault direction'][0]
                if float(rt['modifier']['along'][0]) < 1.0:
                    msg = f"Rocktype {rt_name} has fault modifier 'along' < 1.0"
                    if warn: raise Exception(msg)
                    else: print("Warning: " + msg)
                if float(rt['modifier']['across'][0]) > 1.0:
                    msg = f"Rocktype {rt_name} has fault modifier 'across' > 1.0"
                    if warn: raise Exception(msg)
                    else: print("Warning: " + msg)
                if fault_dir == 1:
                    k_modifier = [
                        float(rt['modifier']['along'][0]),
                        float(rt['modifier']['across'][0]),
                        float(rt['modifier']['vertical'][0]),
                    ]
                elif fault_dir == 2:
                    k_modifier = [
                        float(rt['modifier']['across'][0]),
                        float(rt['modifier']['along'][0]),
                        float(rt['modifier']['vertical'][0]),
                    ]
                else:
                    raise Exception(f"Fault direction {rt['fault direction'][0]} not 1 or 2 in rocktype {rt_name}")
            else:
                k_modifier = [
                    float(rt['modifier']['k1'][0]),
                    float(rt['modifier']['k2'][0]),
                    float(rt['modifier']['k3'][0]),
                ]
            c_modifier = 1.0
            if 'conductivity' in rt['modifier']:
                c_modifier = float(rt['modifier']['conductivity'][0])
            # update values in place
            base_name = rt['base'][0]
            if base_name in rocktype_json:
                rt['permeability'] = [k * m for k, m in zip(
                    rocktype_json[base_name]['permeability'], k_modifier)]
                rt['porosity'][0] = rocktype_json[base_name]['porosity'][0] * rt['modifier']['porosity'][0]
                if 'conductivity' in rocktype_json[base_name]:
                    # print(f"Rocktype {rt_name} use conductivity from base rocktype {base_name}")
                    rt['conductivity'] = [rocktype_json[base_name]['conductivity'][0] * c_modifier]
                else:
                    if 'conductivity' in rt:
                        print(f"(gimu)    Rocktype {rt_name} remove conductivity because no conductivity in base rocktype {base_name}")
                        del rt['conductivity']
            else:
                msg = f"Base rocktype {base_name} not found for rocktype {rt_name}"
                if warn:
                    raise Exception(msg)
                else:
                    print("(gimu)    Warning: " + msg)
                    rt['error'] = msg


def update_waiwera_model_rocktypes(wjson, rtjson, skip_missing=True):
    """ rtjson is loaded from the GMF modifier _rocktypes.json
    """
    for rt in wjson['rock']['types']:
        if rt['name'] not in rtjson:
            if skip_missing:
                print(f"(gimu)    Rocktype {rt['name']} not found in _rocktypes.json... skipped")
                continue
            else:
                raise Exception(f"(gimu)    Rocktype {rt['name']} not found in _rocktypes.json")
        modifier = rtjson[rt['name']]
        rt['permeability'] = modifier['permeability']
        rt['porosity'] = modifier['porosity'][0]
        if 'conductivity' in modifier:
            rt['wet_conductivity'] = modifier['conductivity'][0]
            rt['dry_conductivity'] = modifier['conductivity'][0]

def update_aut2_model_rocktypes(dat, rtjson, skip_missing=True):
    """ rtjson is loaded from the GMF modifier _rocktypes.json
    """
    for rt in dat.grid.rocktypelist:
        if rt.name not in rtjson:
            if skip_missing:
                print(f"(gimu)    Rocktype {rt.name} not found in _rocktypes.json... skipped")
                continue
            else:
                raise Exception(f"(gimu)    Rocktype {rt.name} not found in _rocktypes.json")
        modifier = rtjson[rt.name]
        rt.permeability = modifier['permeability']
        rt.porosity = modifier['porosity'][0]
        if 'conductivity' in modifier:
            rt.conductivity = modifier['conductivity'][0]




""" use: save2incon a.save b.incon [-reset_kcyc] """

from sys import *
from t2incons import *

def main():
    if len(argv) < 2:
        print('use: save2incon a.save b.incon [-reset_kcyc] [-reset_porosity]')
        exit(1)

    readFrom = argv[1]
    saveTo = argv[2]

    if len(argv) > 3:
        opts = argv[3:]
    else:
        opts = []
    inc = t2incon(readFrom)

    for opt in opts:
        if opt == '-reset_kcyc':
            inc.timing['kcyc'] = 1
            inc.timing['iter'] = 1
        if opt == '-reset_porosity':
            inc.porosity = None

    inc.write(saveTo)

if __name__ == '__main__':
    main()



""" Toolbox for accessing GIM/GMF well data

Use:
    from gmf_data import data_json

    data_json.set_data_folder('/path/to/data')

"""

import json
import os
from collections import OrderedDict


class WellData(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def downhole_temperature(self, date_str):
        """ get direct access to data without fluff """
        try:
            raw_data = self['Downhole_Temperature'][date_str]['Data']
        except KeyError as e:
            print(f"Data '{date_str}' does not exist in {self['Well_Name']}")
            print(f"-> available data set are: {list(self['Downhole_Temperature'].keys())}")
            raise e
        elevs, temps = [], []
        for elev_str, temp in raw_data.items():
            elev = float(elev_str)
            elevs.append(elev)
            temps.append(temp)

        # TODO takes care of units in GMF
        return elevs, temps


class DataJSON:
    def __init__(self, max_cache_size=10):
        self._data_folders = []
        self._cache = OrderedDict()
        self._max_cache_size = max_cache_size

    def set_data_folders(self, paths):
        self._data_folders = []
        for path in paths:
            self.add_data_folder(path)

    def add_data_folder(self, path):
        if not os.path.isdir(path):
            raise ValueError(f"Path '{path}' is not a valid directory.")
        if path not in self._data_folders:
            self._data_folders.append(path)

    def remove_data_folder(self, path):
        if path in self._data_folders:
            self._data_folders.remove(path)

    def __getitem__(self, well_name):
        if not self._data_folders:
            raise RuntimeError("Data folder not set. Please call add_data_folder() or set_data_folders() first.")

        if well_name in self._cache:
            self._cache.move_to_end(well_name)
            return self._cache[well_name]

        for folder in self._data_folders:
            file_path = os.path.join(folder, f"{well_name}_data.json")
            if os.path.exists(file_path):
                with open(file_path, 'r') as f:
                    data = WellData(json.load(f))

                if len(self._cache) >= self._max_cache_size:
                    self._cache.popitem(last=False)

                self._cache[well_name] = data
                return data

        msg = '\n'.join([
            f"Well '{well_name}' not found in any of the data folders:",
            ] + [f"    {f}" for f in self._data_folders])
        raise KeyError(msg)

    def __contains__(self, well_name):
        if well_name in self._cache:
            return True

        for folder in self._data_folders:
            file_path = os.path.join(folder, f"{well_name}_data.json")
            if os.path.exists(file_path):
                return True
        return False

    def clear_cache(self):
        self._cache.clear()

    @property
    def data_folders(self):
        return self._data_folders

    @property
    def cached_wells(self):
        return list(self._cache.keys())

    @property
    def wells(self):
        """ Returns a list of all available well names from the data folders.
        """
        well_names = set()
        for folder in self._data_folders:
            for filename in os.listdir(folder):
                if filename.endswith('_data.json'):
                    well_name = filename[:-len('_data.json')]
                    well_names.add(well_name)
        return sorted(list(well_names))


# Singleton instance
data_json = DataJSON()
